package cz.muni.fi.lasaris.sbms.api;

import javax.annotation.security.PermitAll;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.MediaType;

import org.apache.commons.lang3.StringUtils;

import cz.muni.fi.lasaris.sbms.api.entities.DataPoint;
import cz.muni.fi.lasaris.sbms.api.entities.ObservedProperty;
import cz.muni.fi.lasaris.sbms.api.entities.Scope;
import cz.muni.fi.lasaris.sbms.api.request.DataPointsRequest;
import cz.muni.fi.lasaris.sbms.api.response.GroupedResponse;
import cz.muni.fi.lasaris.sbms.model.Fields;
import cz.muni.fi.lasaris.sbms.model.QueryParser;

@Path("influence")
@PermitAll
public class InfluencesEndpoint {

	@GET
	@Path("/influencing")
	@Produces(MediaType.APPLICATION_JSON)
	public GroupedResponse getInfluencing(
		@QueryParam("influenced.scope.bimId") String scopeBimIdP,
		@QueryParam("influenced.property.domain") String propDomP,
		@QueryParam("influenced.property.quality") String propQualityP,
		@QueryParam("grouping") String groupingP,
		@QueryParam("fields") String fieldsP) {
		
		if(scopeBimIdP == null && propDomP == null && propQualityP == null) {
			return GroupedResponse.getErrorResponse("Parameters must be set");
		}
		
		if(!(propDomP == null && propQualityP == null) 
				&& (propDomP == null || propQualityP == null)) {
			return GroupedResponse.getErrorResponse("Both parameters (domain, quality) must be set");
		}
		
		if(propDomP != null && scopeBimIdP == null) {
			return GroupedResponse.getErrorResponse("Scope must be set when property is set");
		}
		
		DataPointsRequest r = new DataPointsRequest();
	
		DataPoint dp = new DataPoint();
		r.setDataPoint(dp);
		
		dp.setOneInfluence();
		if(StringUtils.isNotBlank(scopeBimIdP)) {
			dp.getFirstInfluence().setScope(new Scope());
			dp.getFirstInfluence().getScope().setBimId(scopeBimIdP);
		}
		if(StringUtils.isNotBlank(propDomP)) {
			dp.getFirstInfluence().setProperty(new ObservedProperty());
			dp.getFirstInfluence().getProperty().setDomain(propDomP);
		}
		if(StringUtils.isNotBlank(propQualityP)){
			if(dp.getFirstInfluence().getProperty() == null) {
				dp.getFirstInfluence().setProperty(new ObservedProperty());
			}
			dp.getFirstInfluence().getProperty().setQuality(propQualityP);
		}
	
		r.parseResponseFields(fieldsP);
		r.getResponseFields().addAll(Fields.INFLUENCE_FIELDS);
		r.setGrouping(groupingP);
		return QueryParser.getInfluencingResponse(r);
	}
	
	@GET
	@Path("/influenced")
	@Produces(MediaType.APPLICATION_JSON)
	public GroupedResponse getInfluenced(
			@QueryParam("bmsId") String bmsIdP) {
		
		if(StringUtils.isBlank(bmsIdP)) {
			GroupedResponse.getErrorResponse("bmsId must be set");
		}
		DataPointsRequest r = new DataPointsRequest();
		DataPoint dp = new DataPoint();
		r.setDataPoint(dp);
		
		if(StringUtils.isNotBlank(bmsIdP)) {
			dp.setBmsId(bmsIdP);
		}
		r.getResponseFields().add("bmsId");
		r.getResponseFields().addAll(Fields.INFLUENCE_FIELDS);
		return QueryParser.getInfluencedResponse(r);
	}


// Inserting and Removing of influence is performed by DataPointsEndpoint
	
}
